# epydemix/population/__init__.py

from .population import Population, load_epydemix_population, get_available_locations


__all__ = [
    'Population', 
    'load_epydemix_population',
    'get_available_locations'
]

