# epydemix/__init__.py

from .model.epimodel import EpiModel, simulate
from .model.predefined_models import load_predefined_model
__all__ = [
    'EpiModel',
    'simulate',
    'load_predefined_model'
]
